#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg


od_driver_opts = [
    cfg.StrOpt('odl_username',
               default='admin',
               help=_("OpenDaylight Controller Username")),
    cfg.StrOpt('odl_password',
               default='admin',
               help=_("OpenDaylight Controller Password")),
    cfg.StrOpt('odl_host',
               default='127.0.0.1',
               help=_("OpenDaylight Controller host ip address")),
    cfg.StrOpt('odl_port',
               default='8080',
               help=_("OpenDaylight Controller Rest API port number")),
]


cfg.CONF.register_opts(od_driver_opts, "odl_driver")
